import os
from dataclasses import dataclass
from typing import Any, Iterable, Iterator

import torch

from elasticai.creator.resource_utils import PathType
from elasticai.creator.vhdl.language import Code
from elasticai.creator.vhdl.translator.build_function_mapping import (
    BuildFunctionMapping,
)
from elasticai.creator.vhdl.translator.pytorch.build_function_mappings import (
    DEFAULT_BUILD_FUNCTION_MAPPING,
)


@dataclass
class CodeFile:
    file_name: str
    code: Code


@dataclass
class CodeModule:
    module_name: str
    files: Iterable[CodeFile]


def translate_model(
    model: torch.nn.Module,
    translation_args: dict[str, Any],
    build_function_mapping: BuildFunctionMapping = DEFAULT_BUILD_FUNCTION_MAPPING,
) -> Iterator[CodeModule]:
    """
    Translates a given PyTorch-model to an intermediate representation. The intermediate representation is represented
    as an iterator of VHDLModule objects.

    Parameters:
        model (torch.nn.Module): The PyTorch-model that should be translated.
        build_function_mapping (BuildFunctionMapping):
            Object that maps a given PyTorch-layer to its corresponding build function. If not given the default build
            functions will be used.

    Returns:
        Iterator[CodeModule]:
            Iterator of containers that holds the actual VHDL code. The CodeModule objects holds all CodeFile
            objects for one VHDLModule object (module) and the name of that module. One CodeFile object
            holds the file name of that code files and the actual code as an iterable of str (lines).
    """
    flat_model = filter(
        lambda x: not isinstance(x, (type(model), torch.nn.Sequential)), model.modules()
    )

    for layer_index, layer in enumerate(flat_model):
        layer_class_name = type(layer).__name__

        build_fn = build_function_mapping.get_from_layer(layer)
        if build_fn is None:
            raise NotImplementedError(
                f"Layer '{layer_class_name}' is currently not supported."
            )
        module = build_fn(layer)

        args = translation_args.get(layer_class_name)
        components = module.components(args)
        files = map(lambda x: CodeFile(file_name=x.file_name, code=x()), components)

        yield CodeModule(module_name=f"{layer_index}_{layer_class_name}", files=files)


def save_code(code_repr: Iterable[CodeModule], path: PathType) -> None:
    """
    Saves the generated code on the file system.

    Parameters:
        code_repr (Iterable[CodeModule]): The generated code that should be saved.
        path (PathType):
            The path to a folder in which the code should be saved. All parent folders that don't exist will be created.
    """
    for module in code_repr:
        module_path = os.path.join(path, module.module_name)
        os.makedirs(module_path, exist_ok=True)

        for code_file in module.files:
            file_path = os.path.join(module_path, code_file.file_name)
            code = "\n".join(code_file.code)

            with open(file_path, "w") as out_file:
                out_file.write(code)



from typing import Any

import torch

from elasticai.creator.vhdl.number_representations import (
    FixedPointFactory,
    fixed_point_params_from_factory,
)


class FixedPointQuantFunction(torch.autograd.Function):
    @staticmethod
    def forward(
        ctx: Any, x: torch.Tensor, fixed_point_factory: FixedPointFactory
    ) -> torch.Tensor:
        total_bits, frac_bits = fixed_point_params_from_factory(fixed_point_factory)
        largest_fp_int = 2 ** (total_bits - 1) - 1
        return (
            (x * (1 << frac_bits)).int().float().clamp(-largest_fp_int, largest_fp_int)
        )

    @staticmethod
    def backward(ctx: Any, *grad_outputs: Any) -> Any:
        return *grad_outputs, None


class FixedPointDequantFunction(torch.autograd.Function):
    @staticmethod
    def forward(
        ctx: Any, x: torch.Tensor, fixed_point_factory: FixedPointFactory
    ) -> torch.Tensor:
        total_bits, frac_bits = fixed_point_params_from_factory(fixed_point_factory)
        min_value = 2 ** (total_bits - frac_bits - 1) * (-1)
        max_value = (2 ** (total_bits - 1) - 1) / (1 << frac_bits)
        return (x / (1 << frac_bits)).clamp(min_value, max_value)

    @staticmethod
    def backward(ctx: Any, *grad_outputs: Any) -> Any:
        return *grad_outputs, None



import unittest

import torch
from torch.utils.data import Dataset, TensorDataset

from elasticai.creator.vhdl.evaluators.inference_evaluator import (
    FullPrecisisonInferenceEvaluator,
    QuantizedInferenceEvaluator,
)


class ModelMock(torch.nn.Module):
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return x + 1

    def quantized_forward(self, x: torch.Tensor) -> torch.Tensor:
        return x + 2


def get_dataset(samples: list, labels: list) -> Dataset:
    return TensorDataset(torch.tensor(samples), torch.tensor(labels))


def to_list(x: torch.Tensor) -> list[float]:
    return x.detach().numpy().tolist()


class FullPrecisionInferenceEvaluatorTest(unittest.TestCase):
    def test_full_precision_inference_evaluator_empty_dataset(self) -> None:
        model = ModelMock()
        samples: list[float] = []
        labels: list[float] = []
        evaluator = FullPrecisisonInferenceEvaluator(
            module=model, data=get_dataset(samples, labels)
        )
        predictions, passed_through_labels = evaluator.run()
        self.assertEqual(labels, to_list(passed_through_labels))
        self.assertEqual(labels, to_list(predictions))

    def test_full_precision_inference_evaluator_filled_single_dim_dataset(self) -> None:
        model = ModelMock()
        samples: list[float] = [1, 2, 3]
        labels: list[float] = [2, 3, 4]
        evaluator = FullPrecisisonInferenceEvaluator(
            module=model, data=get_dataset(samples, labels)
        )
        predictions, passed_through_labels = evaluator.run()
        self.assertEqual(labels, to_list(passed_through_labels))
        self.assertEqual(labels, to_list(predictions))

    def test_full_precision_inference_evaluator_filled_multi_dim_dataset(self) -> None:
        model = ModelMock()
        samples: list[list[float]] = [[1, 2], [3, 4], [5, 6]]
        labels: list[list[float]] = [[2, 3], [4, 5], [6, 7]]
        evaluator = FullPrecisisonInferenceEvaluator(
            module=model, data=get_dataset(samples, labels)
        )
        predictions, passed_through_labels = evaluator.run()
        self.assertEqual(labels, to_list(passed_through_labels))
        self.assertEqual(labels, to_list(predictions))


class QuantizedInferenceEvaluatorTest(unittest.TestCase):
    def test_quantized_inference_evaluator_empty_dataset(self) -> None:
        model = ModelMock()
        samples: list[float] = []
        labels: list[float] = []
        evaluator = QuantizedInferenceEvaluator(
            module=model, data=get_dataset(samples, labels)
        )
        predictions, passed_through_labels = evaluator.run()
        self.assertEqual(labels, to_list(passed_through_labels))
        self.assertEqual(labels, to_list(predictions))

    def test_quantized_inference_evaluator_no_input_output_quantizers(self) -> None:
        model = ModelMock()
        samples: list[list[float]] = [[1, 2], [3, 4], [5, 6]]
        labels: list[list[float]] = [[3, 4], [5, 6], [7, 8]]
        evaluator = QuantizedInferenceEvaluator(
            module=model, data=get_dataset(samples, labels)
        )
        predictions, passed_through_labels = evaluator.run()
        self.assertEqual(labels, to_list(passed_through_labels))
        self.assertEqual(labels, to_list(predictions))

    def test_quantized_inference_evaluator_input_output_quantizers(self) -> None:
        model = ModelMock()
        samples: list[list[float]] = [[1, 2], [3, 4], [5, 6]]
        labels: list[list[float]] = [[2, 3], [4, 5], [6, 7]]
        evaluator = QuantizedInferenceEvaluator(
            module=model,
            data=get_dataset(samples, labels),
            input_quant=lambda x: x * 2,
            output_quant=lambda x: x / 2,
        )
        predictions, passed_through_labels = evaluator.run()
        self.assertEqual(labels, to_list(passed_through_labels))
        self.assertEqual(labels, to_list(predictions))



import torch
import torch.nn.functional as F

from elasticai.creator.vhdl.number_representations import FixedPointFactory
from elasticai.creator.vhdl.quantized_modules.autograd_functions import (
    FixedPointDequantFunction,
    FixedPointQuantFunction,
)
from elasticai.creator.vhdl.quantized_modules.typing import QuantType


class _ReLUBase(torch.nn.ReLU):
    def __init__(
        self,
        input_quant: QuantType = lambda x: x,
        input_dequant: QuantType = lambda x: x,
        inplace: bool = False,
    ) -> None:
        super().__init__(inplace)
        self.input_quant = input_quant
        self.input_dequant = input_dequant

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        q_x = self.input_dequant(self.input_quant(x))
        return F.relu(q_x)

    def quantized_forward(self, x: torch.Tensor) -> torch.Tensor:
        raise NotImplementedError("The quantized_forward function is not implemented.")


class FixedPointReLU(_ReLUBase):
    def __init__(
        self, fixed_point_factory: FixedPointFactory, inplace: bool = False
    ) -> None:
        super().__init__(
            input_quant=lambda x: FixedPointQuantFunction.apply(x, fixed_point_factory),
            input_dequant=lambda x: FixedPointDequantFunction.apply(
                x, fixed_point_factory
            ),
            inplace=inplace,
        )

    def quantized_forward(self, x: torch.Tensor) -> torch.Tensor:
        return F.relu(x)



from collections.abc import Iterator

import torch

from elasticai.creator.vhdl.translator.abstract.layers import LSTMModule


def build_lstm(lstm: torch.nn.LSTM) -> LSTMModule:
    def to_list(tensor: torch.Tensor) -> list:
        return tensor.detach().numpy().tolist()

    def get_weights(weight_prefix: str) -> Iterator[list]:
        for i in range(lstm.num_layers):
            yield to_list(getattr(lstm, f"{weight_prefix}_l{i}"))

    return LSTMModule(
        weights_ih=list(get_weights("weight_ih")),
        weights_hh=list(get_weights("weight_hh")),
        biases_ih=list(get_weights("bias_ih")),
        biases_hh=list(get_weights("bias_hh")),
    )



from elasticai.creator.vhdl.components.dual_port_2_clock_ram_component import (
    DualPort2ClockRamComponent,
)
from elasticai.creator.vhdl.components.linear_1d_component import Linear1dComponent
from elasticai.creator.vhdl.components.lstm_common_component import LSTMCommonComponent
from elasticai.creator.vhdl.components.lstm_component import LSTMComponent
from elasticai.creator.vhdl.components.rom_component import RomComponent
from elasticai.creator.vhdl.components.sigmoid_component import SigmoidComponent
from elasticai.creator.vhdl.components.tanh_component import TanhComponent



import unittest
from difflib import diff_bytes

import torch
import torch.nn.functional as F

from elasticai.creator.vhdl.number_representations import (
    FixedPoint,
    float_values_to_fixed_point,
)
from elasticai.creator.vhdl.quantized_modules.relu import FixedPointReLU, _ReLUBase


def to_list(x: torch.Tensor) -> list[float]:
    return x.detach().numpy().tolist()


def to_fixed_point(values: list[float], total_bits: int, frac_bits: int) -> list[int]:
    def to_fp(value: FixedPoint) -> int:
        return value.to_signed_int()

    return list(map(to_fp, float_values_to_fixed_point(values, total_bits, frac_bits)))


def from_fixed_point(values: list[int], total_bits: int, frac_bits: int) -> list[float]:
    def to_float(value: int) -> float:
        return float(FixedPoint.from_signed_int(value, total_bits, frac_bits))

    return list(map(to_float, values))


class ReLUBaseTest(unittest.TestCase):
    def test_relu_base_not_quantized_behaves_equivalent_to_relu(self) -> None:
        relu = _ReLUBase()

        xs = torch.linspace(-10, 10, 100)
        expected = to_list(F.relu(xs))
        actual = to_list(relu(xs))

        self.assertEquals(expected, actual)

    def test_relu_base_quantized_behaves_equivalent_to_relu(self) -> None:
        relu = _ReLUBase(input_quant=lambda x: x + 2, input_dequant=lambda x: x - 1)

        xs = torch.arange(-10, 10)
        expected = to_list(F.relu(xs + 1))
        actual = to_list(relu(xs))

        self.assertEquals(expected, actual)

    def test_relu_base_quantized_forward_raises_error(self) -> None:
        with self.assertRaises(NotImplementedError):
            _ReLUBase().quantized_forward(torch.ones(1))


class FixedPointReLUTest(unittest.TestCase):
    def test_fixed_point_relu_forward_works_as_expected(self) -> None:
        fp_args = 8, 3
        relu = FixedPointReLU(fixed_point_factory=FixedPoint.get_factory(*fp_args))

        xs = torch.linspace(-10, 10, 100)
        fake_quantized_xs = from_fixed_point(
            to_fixed_point(to_list(xs), *fp_args), *fp_args
        )
        fake_quantized_xs = torch.tensor(fake_quantized_xs, dtype=torch.float32)
        expected = to_list(F.relu(fake_quantized_xs))
        actual = to_list(relu(xs))

        self.assertEquals(expected, actual)

    def test_fixed_point_quantized_forward_works_as_expected(self) -> None:
        relu = FixedPointReLU(
            fixed_point_factory=FixedPoint.get_factory(total_bits=8, frac_bits=3)
        )

        xs = torch.linspace(-10, 10, 20)
        expected = to_list(F.relu(xs))
        actual = to_list(relu.quantized_forward(xs))

        self.assertEquals(expected, actual)



import unittest

import torch
from torch.nn.parameter import Parameter

from elasticai.creator.vhdl.number_representations import FixedPoint
from elasticai.creator.vhdl.quantized_modules.linear import (
    FixedPointLinear,
    _LinearBase,
)


def to_list(x: torch.Tensor) -> list[float]:
    return x.detach().numpy().tolist()


class LinearBaseTest(unittest.TestCase):
    def test_default_linear(self) -> None:
        linear = _LinearBase(in_features=3, out_features=1)
        linear.weight = Parameter(torch.ones_like(linear.weight))
        linear.bias = Parameter(torch.ones_like(linear.bias) * 4)

        input_tensor = torch.as_tensor([1, 2, 3], dtype=torch.float32)
        actual = to_list(linear(input_tensor))
        target = [10]

        self.assertEqual(actual, target)

    def test_linear_with_customized_add_op(self) -> None:
        linear = _LinearBase(in_features=3, out_features=1, add_op=torch.sub)
        linear.weight = Parameter(torch.ones_like(linear.weight))
        linear.bias = Parameter(torch.ones_like(linear.bias) * 4)

        input_tensor = torch.as_tensor([1, 2, 3], dtype=torch.float32)
        actual = to_list(linear(input_tensor))
        target = [2]

        self.assertEqual(actual, target)

    def test_linear_with_customized_matmul_op(self) -> None:
        def custom_matmul(a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
            return torch.matmul(a, b) * (-1)

        linear = _LinearBase(in_features=3, out_features=1, matmul_op=custom_matmul)
        linear.weight = Parameter(torch.ones_like(linear.weight))
        linear.bias = Parameter(torch.ones_like(linear.bias) * 4)

        input_tensor = torch.as_tensor([1, 2, 3], dtype=torch.float32)
        actual = to_list(linear(input_tensor))
        target = [-2]

        self.assertEqual(actual, target)

    def test_linear_with_customized_add_op_and_matmul_op(self) -> None:
        def custom_matmul(a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
            return torch.matmul(a, b) * (-1)

        linear = _LinearBase(
            in_features=3, out_features=1, matmul_op=custom_matmul, add_op=torch.sub
        )
        linear.weight = Parameter(torch.ones_like(linear.weight))
        linear.bias = Parameter(torch.ones_like(linear.bias) * 4)

        input_tensor = torch.as_tensor([1, 2, 3], dtype=torch.float32)
        actual = to_list(linear(input_tensor))
        target = [-10]

        self.assertEqual(actual, target)

    def test_linear_with_customized_matmul_without_bias(self) -> None:
        def custom_matmul(a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
            return torch.matmul(a, b) * (-1)

        linear = _LinearBase(
            in_features=3, out_features=1, matmul_op=custom_matmul, bias=False
        )
        linear.weight = Parameter(torch.ones_like(linear.weight))

        input_tensor = torch.as_tensor([1, 2, 3], dtype=torch.float32)
        actual = to_list(linear(input_tensor))
        target = [-6]

        self.assertEqual(actual, target)

    def test_linear_quantized_forward_raises_error(self) -> None:
        with self.assertRaises(NotImplementedError):
            _LinearBase(1, 2).quantized_forward(torch.ones(1))


class FixedPointLinearTest(unittest.TestCase):
    def test_fixed_point_linear_in_bounds(self) -> None:
        fp_factory = FixedPoint.get_factory(total_bits=16, frac_bits=8)
        linear = FixedPointLinear(
            in_features=3, out_features=1, fixed_point_factory=fp_factory
        )
        linear.weight = Parameter(torch.ones_like(linear.weight))
        linear.bias = Parameter(torch.ones_like(linear.bias))

        input_tensor = torch.as_tensor([-7, 2, 3], dtype=torch.float32)
        actual = to_list(linear(input_tensor))
        target = [-1]

        self.assertEqual(actual, target)

    def test_fixed_point_linear_out_of_bounds(self) -> None:
        fp_factory = FixedPoint.get_factory(total_bits=3, frac_bits=0)
        linear = FixedPointLinear(
            in_features=3, out_features=1, fixed_point_factory=fp_factory
        )
        linear.weight = Parameter(torch.ones_like(linear.weight))
        linear.bias = Parameter(torch.ones_like(linear.bias))

        input_tensor = torch.as_tensor([1, 2, 3], dtype=torch.float32)
        actual = to_list(linear(input_tensor))
        target = [7]

        self.assertEqual(actual, target)

    def test_fixed_point_linear_quantized_forward(self) -> None:
        fp_factory = FixedPoint.get_factory(total_bits=8, frac_bits=4)
        linear = FixedPointLinear(
            in_features=3, out_features=1, fixed_point_factory=fp_factory
        )
        linear.weight = Parameter(torch.ones_like(linear.weight))
        linear.bias = Parameter(torch.ones_like(linear.bias))

        input_tensor = torch.tensor(
            [fp_factory(x).to_signed_int() for x in [-1, 0.5, 2]], dtype=torch.float32
        )
        expected = [40]
        actual = to_list(linear.quantized_forward(input_tensor))

        self.assertEquals(expected, actual)



import unittest
from dataclasses import dataclass
from typing import Any

import elasticai.creator.qat.layers as qtorch
from elasticai.creator.vhdl.language import Code
from elasticai.creator.vhdl.translator.build_function_mapping import (
    BuildFunctionMapping,
)
from elasticai.creator.vhdl.vhdl_component import VHDLComponent, VHDLModule


class MockVHDLComponent(VHDLComponent):
    @property
    def file_name(self) -> str:
        return "test_component"

    def __call__(self) -> Code:
        return ["line1", "line2"]


@dataclass
class MockModule(VHDLModule):
    vhdl_components: list[VHDLComponent]

    def components(self, args: Any) -> list[VHDLComponent]:
        return self.vhdl_components


def mock_build_function1(layer: Any) -> VHDLModule:
    return MockModule([MockVHDLComponent()])


def mock_build_function2(layer: Any) -> VHDLModule:
    return MockModule([MockVHDLComponent()])


def mock_build_function3(layer: Any) -> VHDLModule:
    return MockModule([MockVHDLComponent()])


class BuildFunctionMappingTest(unittest.TestCase):
    def setUp(self) -> None:
        self.linear_layer = qtorch.QLinear(1, 1, qtorch.Binarize())
        self.conv1d_layer = qtorch.QConv1d(1, 1, (1,), qtorch.Binarize())
        self.conv2d_layer = qtorch.QConv2d(1, 1, (1, 1), qtorch.Binarize())

        self.mapping_dict = {
            "torch.nn.utils.parametrize.ParametrizedQLinear": mock_build_function1,
            "torch.nn.utils.parametrize.ParametrizedQConv1d": mock_build_function2,
        }

        self.mapping = BuildFunctionMapping(mapping=self.mapping_dict)

    def test_get_from_layer_existing_key(self) -> None:
        self.assertEqual(
            mock_build_function1, self.mapping.get_from_layer(self.linear_layer)
        )

    def test_get_from_layer_not_existing_key(self) -> None:
        self.assertEqual(None, self.mapping.get_from_layer(self.conv2d_layer))

    def test_to_dict(self) -> None:
        self.assertEqual(self.mapping_dict, self.mapping.to_dict())

    def test_join_with_dict_new_and_existing_keys(self) -> None:
        new_mapping = self.mapping.join_with_dict(
            {
                "torch.nn.utils.parametrize.ParametrizedQConv2d": mock_build_function3,
                "torch.nn.utils.parametrize.ParametrizedQConv1d": mock_build_function1,
            }
        )
        target_mapping_dict = {
            "torch.nn.utils.parametrize.ParametrizedQLinear": mock_build_function1,
            "torch.nn.utils.parametrize.ParametrizedQConv1d": mock_build_function1,
            "torch.nn.utils.parametrize.ParametrizedQConv2d": mock_build_function3,
        }
        self.assertEqual(target_mapping_dict, new_mapping.to_dict())



import unittest

from elasticai.creator.vhdl.components import RomComponent
from elasticai.creator.vhdl.number_representations import FixedPoint


class RomComponentTest(unittest.TestCase):
    def setUp(self) -> None:
        fp = FixedPoint.get_factory(total_bits=16, frac_bits=8)
        self.rom = RomComponent(
            rom_name="test_rom",
            values=[fp(i) for i in range(20)],
            resource_option="auto",
        )

    def test_data_width_correct_derived(self) -> None:
        self.assertEqual(self.rom.data_width, 16)

    def test_addr_width_correct_calculated(self) -> None:
        self.assertEqual(self.rom.addr_width, 5)

    def test_correct_number_of_values(self) -> None:
        self.assertEqual(len(self.rom.hex_values), 32)

    def test_values_correct_padded(self) -> None:
        self.assertEqual(['x"0000"'] * 12, self.rom.hex_values[20:])



import unittest

from elasticai.creator.vhdl.components import Linear1dComponent
from elasticai.creator.vhdl.number_representations import FixedPoint


class Linear1dComponentTest(unittest.TestCase):
    def setUp(self) -> None:
        self.component = Linear1dComponent(
            in_features=20,
            out_features=1,
            fixed_point_factory=FixedPoint.get_factory(total_bits=16, frac_bits=8),
            work_library_name="work",
        )

    def test_derives_correct_data_width(self) -> None:
        self.assertEqual(self.component.data_width, 16)

    def test_calculates_correct_addr_width(self) -> None:
        self.assertEqual(self.component.addr_width, 5)

    def test_out_features_larger_1_raises_not_implemented_error(self) -> None:
        with self.assertRaises(NotImplementedError):
            _ = Linear1dComponent(
                in_features=3,
                out_features=2,
                fixed_point_factory=FixedPoint.get_factory(total_bits=8, frac_bits=4),
            )
